#include <QStringListModel>
#include <QList>
#include <QPointer>
#include <QFile>
#include <QTextStream>
#include <QMessageBox>
#include <QString>
#include <QRegExp>
#include <QDomDocument>
#include <QDomElement>
#include <QDomText>
#include <QDomNodeList>

#include "const.h"
#include "dictionary.h"
#include "phonologypage.h"
#include "phonotacticspage.h"
#include "suprasegmentalspage.h"
#include "wordpage.h"
#include "searchpage.h"
#include "dictionarymanagementpage.h"
#include "phoneme.h"
#include "phonotactics.h"
#include "word.h"
#include "syllable.h"
#include "loadpartialwordlistdialog.h"

Dictionary::Dictionary ()  {
  setMinimumWidth (700);
  setMinimumHeight (550);

  searchPage = new SearchPage;
  phonologyPage = new PhonologyPage;
  phonotacticsPage = new PhonotacticsPage;
  suprasegmentalsPage = new SuprasegmentalsPage;
  wordPage = new WordPage;
  dictionaryManagementPage = new DictionaryManagementPage;

  languageName = "Conlang";

  loadPartialDialog = NULL;

  addTab (searchPage, "Search");
  addTab (phonologyPage, "Phonemes");
  addTab (suprasegmentalsPage, "Suprasegmentals");
  addTab (phonotacticsPage, "Phonotactics");
  addTab (wordPage, "Words");
  addTab (dictionaryManagementPage, "Load/Save Dictionaries");

  connect (phonologyPage,
           SIGNAL (phonemeListUpdated (QList< QPointer<Phoneme> >)),
           wordPage, SLOT (setPhonemeList (QList< QPointer<Phoneme> >)));
  connect (phonologyPage,
           SIGNAL (alphabetUpdated (QList< QPointer<Phoneme> >)),
           wordPage, SLOT (setAlphabetOrder (QList< QPointer<Phoneme> >)));
  connect (phonologyPage,
           SIGNAL (phonemeListUpdated (QList< QPointer<Phoneme> >)),
           suprasegmentalsPage,
           SLOT (setPhonemeList (QList< QPointer<Phoneme> >)));
  connect (phonologyPage,
           SIGNAL (phonemeListUpdated (QList< QPointer<Phoneme> >)),
           phonotacticsPage, SLOT (setPhonemeList (QList< QPointer<Phoneme> >)));
  connect (phonotacticsPage, SIGNAL (phonotacticsUpdated (Phonotactics, bool)),
           wordPage, SLOT (setPhonotactics (Phonotactics, bool)));
  connect (suprasegmentalsPage,
           SIGNAL (suprasegmentalListUpdated(QList<QPointer<Suprasegmental> >)),
           wordPage,
           SLOT (setSuprasegmentalList (QList< QPointer<Suprasegmental> >)));
  connect (wordPage, SIGNAL (wordListUpdated (QList< QPointer<Word> >)),
           searchPage, SLOT (setWordList (QList< QPointer<Word> >)));

  connect (dictionaryManagementPage, SIGNAL (loadFrom (QString, int)), this,
           SLOT (load (QString, int)));
  connect (dictionaryManagementPage, SIGNAL (loadPartialWordlist (QDomElement)), this,
           SLOT (loadPartialWordlist (QDomElement)));
  connect (dictionaryManagementPage, SIGNAL (loadTextFrom (QString, QString, bool)),
           this, SLOT (loadText (QString, QString, bool)));
  connect (dictionaryManagementPage, SIGNAL (loadLexiqueFrom (QString, QStringList, QList<int>, bool)),
           this, SLOT (loadLexique (QString, QStringList, QList<int>, bool)));
  connect (dictionaryManagementPage, SIGNAL (saveAs (QString)), this,
           SLOT (save (QString)));
  connect (dictionaryManagementPage, SIGNAL (saveTextAs (QString, QString, bool)),
           this, SLOT (saveText (QString, QString, bool)));
  connect (dictionaryManagementPage, SIGNAL (setLanguageName (QString)), this,
           SLOT (setLanguageName (QString)));
  connect (dictionaryManagementPage, SIGNAL (clearLists ()), phonologyPage,
           SLOT (clear ()));
  connect (dictionaryManagementPage, SIGNAL (clearLists ()), wordPage,
           SLOT (clear ()));
  connect (dictionaryManagementPage, SIGNAL (clearLists ()), searchPage,
           SLOT (clear ()));
  connect (dictionaryManagementPage, SIGNAL (clearLists ()), suprasegmentalsPage,
           SLOT (clear ()));
  connect (dictionaryManagementPage, SIGNAL (clearLists ()), phonotacticsPage,
           SLOT (clear ()));
}

void Dictionary::setPhonemeModel (QStringListModel *model)  {
  phonologyPage->setModel (model);
}

void Dictionary::setWordModel (QStringListModel *model)  {
  wordPage->setModel (model);
}

void Dictionary::setSuprasegmentalModel (QStringListModel *model)  {
  suprasegmentalsPage->setModel (model);
}

void Dictionary::load (QString filename, int version)  {
  if (version == V0_1)  {
    loadV0_1 (filename);
    return;
  }

  QDomDocument doc ("ConlangML");
  QFile file (filename + ".cdic");

  if (!file.open (QFile::ReadOnly))  {
    QMessageBox::warning (this, "", "Cannot read file.");
    return;
  }

  if (!doc.setContent (&file))  {
    QMessageBox::warning (this, "", "Cannot read XML.");
    file.close ();
    return;
  }

  file.close ();

  phonologyPage->clear ();
  wordPage->clear ();
  phonotacticsPage->clear ();
  searchPage->clear ();
  suprasegmentalsPage->clear ();

  languageName = doc.documentElement ().attribute ("name");
  if (languageName != "")
    emit languageNameUpdated (languageName);

  if (!doc.documentElement ().firstChildElement ("squareBrackets").isNull ())  {
    phonologyPage->setSquareBrackets (true);
    wordPage->setSquareBrackets (true);
  }

  loadInventory (doc.documentElement ().firstChildElement ("inventory"));
  loadPhonotactics (doc.documentElement ().firstChildElement ("phonotactics"));
  loadSupras (doc.documentElement ().firstChildElement ("suprasegmentals"));
  loadWordList (doc.documentElement ().firstChildElement ("wordList"));
}

void Dictionary::loadInventory (QDomElement inv)  {
  if (inv.isNull ()) return;

  QList< QPointer<Phoneme> > phonemeList;
  QString alphabet = inv.firstChildElement ("alphabet").text ();
  QStringList alphaStrings = alphabet.split (" ");

  QDomNodeList nodeList = inv.childNodes ();

  for (int x = 0; x < nodeList.size (); x++)  {
    QDomElement currentElement = nodeList.at(x).toElement ();
    if (currentElement.isNull ()) continue;

    if (currentElement.tagName () == "phonemeDef")  {
      phonemeList.append (new Phoneme(currentElement.attribute ("name")));
      phonemeList.last ()->setSpellings
                    (currentElement.firstChildElement ("orthography").text ());
      phonemeList.last ()->setNotes
                    (currentElement.firstChildElement ("notes").text ());
    }
  }

  QList< QPointer<Phoneme> > alphaList;
  for (int x = 0; x < alphaStrings.size (); x++)
    for (int y = 0; y < phonemeList.size (); y++)
      if (alphaStrings[x] == phonemeList[y]->getName ())  {
        alphaList.append (phonemeList[y]);
        break;
      }

  phonologyPage->setList (phonemeList);
  suprasegmentalsPage->setPhonemeList (phonemeList);

  phonologyPage->setAlphabeticList (alphaList);
  wordPage->setAlphabetOrder (alphaList);
}

void Dictionary::loadPhonotactics (QDomElement tax)  {
  if (tax.isNull ()) return;

  Phonotactics phonotax;
  phonotax.setOnsetRequired (tax.attribute ("onsetReq", "0").toInt ());
  phonotax.setPhonemeList (phonologyPage->getList ());

  QList< QList< QPointer<Phoneme> > > legalOnsets;
  QList< QList< QPointer<Phoneme> > > legalPeaks;
  QList< QList< QPointer<Phoneme> > > legalCodas;
  QList< QPointer<Phoneme> > phonemeList = phonologyPage->getList ();

  QDomElement onsetTag = tax.firstChildElement ("onsets");
  QDomElement peakTag = tax.firstChildElement ("peaks");
  QDomElement codaTag = tax.firstChildElement ("codas");
  QDomElement ignoreTag = tax.firstChildElement ("ignore");

  if (!onsetTag.isNull ())  {
    QDomNodeList onsetList = onsetTag.childNodes ();

    for (int x = 0; x < onsetList.size (); x++)  {
      QDomElement currentElement = onsetList.at(x).toElement ();
      if (currentElement.isNull ()) continue;

      if (currentElement.tagName () == "sequence")  {
        QStringList sequenceList = currentElement.text ().split (" ");
        legalOnsets.append (QList< QPointer<Phoneme> > ());

        for (int s = 0; s < sequenceList.size (); s++)
          for (int p = 0; p < phonemeList.size (); p++)
            if (sequenceList[s] == phonemeList[p]->getName ())  {
              legalOnsets.last ().append (phonemeList[p]);
              break;
            }
      }
    }
  }

  phonotax.setOnsetList (legalOnsets);

  if (!peakTag.isNull ())  {
    QDomNodeList peakList = peakTag.childNodes ();

    for (int x = 0; x < peakList.size (); x++)  {
      QDomElement currentElement = peakList.at(x).toElement ();
      if (currentElement.isNull ()) continue;

      if (currentElement.tagName () == "sequence")  {
        QStringList sequenceList = currentElement.text ().split (" ");
        legalPeaks.append (QList< QPointer<Phoneme> > ());

        for (int s = 0; s < sequenceList.size (); s++)
          for (int p = 0; p < phonemeList.size (); p++)
            if (sequenceList[s] == phonemeList[p]->getName ())  {
              legalPeaks.last ().append (phonemeList[p]);
              break;
            }
      }
    }
  }

  phonotax.setPeakList (legalPeaks);

  if (!codaTag.isNull ())  {
    QDomNodeList codaList = codaTag.childNodes ();

    for (int x = 0; x < codaList.size (); x++)  {
      QDomElement currentElement = codaList.at(x).toElement ();
      if (currentElement.isNull ()) continue;

      if (currentElement.tagName () == "sequence")  {
        QStringList sequenceList = currentElement.text ().split (" ");
        legalCodas.append (QList< QPointer<Phoneme> > ());

        for (int s = 0; s < sequenceList.size (); s++)
          for (int p = 0; p < phonemeList.size (); p++)
            if (sequenceList[s] == phonemeList[p]->getName ())  {
              legalCodas.last ().append (phonemeList[p]);
              break;
            }
      }
    }
  }

  if (!ignoreTag.isNull ())
    phonotax.setIgnored (ignoreTag.text ());

  phonotax.setCodaList (legalCodas);

  phonotacticsPage->setPhonotactics (phonotax);
  wordPage->setPhonotactics (phonotax);
}

void Dictionary::loadSupras (QDomElement supras)  {
  if (supras.isNull ()) return;

  QList< QPointer<Suprasegmental> > supraList;
  QList< QPointer<Phoneme> > phonemeList = phonologyPage->getList ();

  QDomNodeList nodeList = supras.childNodes ();
  for (int x = 0; x < nodeList.size (); x++)  {
    QDomElement currentElement = nodeList.at(x).toElement ();
    if (currentElement.isNull ()) continue;

    if (currentElement.tagName () == "supraDef")  {
      supraList.append (new Suprasegmental);
      supraList.last ()->setName (currentElement.attribute ("name"));
      supraList.last ()->setDomain (currentElement.attribute ("domain", "0").toInt ());

      QStringList phonNames = currentElement.firstChildElement ("applies").text ().split (" ");
      QList< QPointer<Phoneme> > appList;
      for (int p = 0; p < phonemeList.size (); p++)
        if (phonNames.contains (phonemeList[p]->getName ()))
          appList.append (phonemeList[p]);

      supraList.last ()->setPhonemeList (appList);

      QDomElement spelling = currentElement.firstChildElement ("spelling");
      supraList.last ()->setSpellingRules (spelling.attribute ("type", "0").toInt (),
                                           spelling.text ());

      QDomElement rep = currentElement.firstChildElement ("representation");
      supraList.last ()->setRepresentationRules (rep.attribute ("type", "0").toInt (),
                                                 rep.text ());

      supraList.last ()->setDescription (currentElement.firstChildElement ("notes").text ());
    }
  }

  suprasegmentalsPage->setList (supraList);
  wordPage->setSuprasegmentalList (supraList);
}

void Dictionary::loadWordList (QDomElement words)  {
  if (words.isNull ()) return;

  QList< QPointer<Word> > wordList;
  Phonotactics phonotax = wordPage->getPhonotactics ();

  QDomNodeList nodeList = words.childNodes ();
  for (int x = 0; x < nodeList.size (); x++)  {
    QDomElement currentElement = nodeList.at(x).toElement ();
    if (currentElement.isNull ()) continue;

    if (currentElement.tagName () == "wordDef")  {
      wordList.append (new Word (currentElement.attribute ("name"), phonotax));

      wordList.last ()->setType (currentElement.attribute ("type"));
      wordList.last ()->setSubType (currentElement.attribute ("subtype"));
      wordList.last ()->setDefinition (currentElement.firstChildElement ("definition").text ());

      wordList.last ()->setPhonology (loadWordPhonology (currentElement.firstChildElement ("phonology")));
    }
  }

  wordPage->setWordList (wordList);
  searchPage->setWordList (wordList);
}

QList<Syllable> Dictionary::loadWordPhonology (QDomElement phonologyElem)  {
  if (phonologyElem.isNull ()) return QList<Syllable> ();

  QList<Syllable> phonology;
  QList< QPointer<Phoneme> > phonemeList = phonologyPage->getList ();
  QList< QPointer<Suprasegmental> > supraList = suprasegmentalsPage->getSupraList ();

  QDomNodeList syllList = phonologyElem.childNodes ();
  for (int s = 0; s < syllList.size (); s++)  {
    QDomElement currentSyll = syllList.at(s).toElement ();
    if (currentSyll.isNull ()) continue;
    if (currentSyll.tagName () != "syllable") continue;

    Syllable syll;

    for (int sup = 0; sup < supraList.size (); sup++)
      if (supraList[sup]->getName () == currentSyll.attribute ("bSupra") ||
          supraList[sup]->getName () == currentSyll.attribute ("fSupra") ||
          supraList[sup]->getName () == currentSyll.attribute ("pSupra"))
        syll.addSupra (supraList[sup]);

    QDomNodeList onsetNodes = currentSyll.firstChildElement ("onset").childNodes ();
    QDomNodeList peakNodes = currentSyll.firstChildElement ("peak").childNodes ();
    QDomNodeList codaNodes = currentSyll.firstChildElement ("coda").childNodes ();

    QList< QPointer<Phoneme> > onsetPhonList;
    QList< QList< QPointer<Suprasegmental> > > onsetSupraList;
    QList< QPointer<Phoneme> > peakPhonList;
    QList< QList< QPointer<Suprasegmental> > > peakSupraList;
    QList< QPointer<Phoneme> > codaPhonList;
    QList< QList< QPointer<Suprasegmental> > > codaSupraList;

    for (int o = 0; o < onsetNodes.size (); o++)  {
      QDomElement onsetElem = onsetNodes.at(o).toElement ();
      if (onsetElem.isNull ()) continue;
      if (onsetElem.tagName () != "phoneme") continue;

      for (int ph = 0; ph < phonemeList.size (); ph++)
        if (phonemeList[ph]->getName () == onsetElem.attribute ("name"))  {
          onsetPhonList.append (phonemeList[ph]);
          break;
        }

      onsetSupraList.append (QList< QPointer<Suprasegmental> > ());

      QDomNodeList supraNodeList = onsetElem.childNodes ();
      for (int sn = 0; sn < supraNodeList.size (); sn++)
        if (!supraNodeList.at(sn).toElement ().isNull () &&
            supraNodeList.at(sn).toElement ().tagName () == "supra")
          for (int sup = 0; sup < supraList.size (); sup++)
            if (supraList[sup]->getName () == supraNodeList.at(sn).toElement ().text ())
              onsetSupraList.last ().append (supraList[sup]);
    }

    for (int p = 0; p < peakNodes.size (); p++)  {
      QDomElement peakElem = peakNodes.at(p).toElement ();
      if (peakElem.isNull ()) continue;
      if (peakElem.tagName () != "phoneme") continue;

      for (int ph = 0; ph < phonemeList.size (); ph++)
        if (phonemeList[ph]->getName () == peakElem.attribute ("name"))  {
          peakPhonList.append (phonemeList[ph]);
          break;
        }

      peakSupraList.append (QList< QPointer<Suprasegmental> > ());

      QDomNodeList supraNodeList = peakElem.childNodes ();
      for (int sn = 0; sn < supraNodeList.size (); sn++)
        if (!supraNodeList.at(sn).toElement ().isNull () &&
            supraNodeList.at(sn).toElement ().tagName () == "supra")
          for (int sup = 0; sup < supraList.size (); sup++)
            if (supraList[sup]->getName () == supraNodeList.at(sn).toElement ().text ())
              peakSupraList.last ().append (supraList[sup]);
    }

    for (int c = 0; c < codaNodes.size (); c++)  {
      QDomElement codaElem = codaNodes.at(c).toElement ();
      if (codaElem.isNull ()) continue;
      if (codaElem.tagName () != "phoneme") continue;

      for (int ph = 0; ph < phonemeList.size (); ph++)
        if (phonemeList[ph]->getName () == codaElem.attribute ("name"))  {
          codaPhonList.append (phonemeList[ph]);
          break;
        }

      codaSupraList.append (QList< QPointer<Suprasegmental> > ());

      QDomNodeList supraNodeList = codaElem.childNodes ();
      for (int sn = 0; sn < supraNodeList.size (); sn++)
        if (!supraNodeList.at(sn).toElement ().isNull () &&
            supraNodeList.at(sn).toElement ().tagName () == "supra")
          for (int sup = 0; sup < supraList.size (); sup++)
            if (supraList[sup]->getName () == supraNodeList.at(sn).toElement ().text ())
              codaSupraList.last ().append (supraList[sup]);
    }

    syll.setOnset (onsetPhonList, onsetSupraList);
    syll.setPeak (peakPhonList, peakSupraList);
    syll.setCoda (codaPhonList, codaSupraList);

    phonology.append (syll);
  }

  return phonology;
}

void Dictionary::loadV0_1 (QString filename)  {
  QFile file (filename + ".cdic");
  QTextStream in (&file);

  if (!file.open (QFile::ReadOnly | QFile::Text))  {
    QMessageBox::warning (this, "", "Cannot read file.");
    return;
  }

  languageName = "Conlang";
  emit languageNameUpdated ("Conlang");

  QList< QPointer<Phoneme> > phonemeList;
  QList< QPointer<Suprasegmental> > suprasegmentalList;

  QString line = in.readLine ();

  while (line != "")  {
    QStringList split = line.split ("  ");
    Phoneme *phoneme = new Phoneme (split[0]);
    phoneme->setSpellings (split[1]);

    QString desc = "";
    for (int x = 2; x < split.size (); x++)
      desc += split[x] + "  ";

    desc.chop (2);
    phoneme->setNotes (desc.replace (";;", "\n"));
    phonemeList.append (phoneme);

    line = in.readLine ();
  }

  QList< QPointer<Word> > wordList;

  line = in.readLine ();

  while (line != "")  {
    Word *word = new Word (line.left (line.indexOf ("[")), Phonotactics ());
    line.remove (0, line.indexOf ("[") + 1);

    QList< QPointer<Phoneme> > wordPhons;
    QStringList phons = line.left (line.indexOf ("]")).split (' ');

    for (int x = 0; x < phons.size (); x++)
      for (int y = 0; y < phonemeList.size (); y++)
        if (phonemeList[y]->getName () == phons[x])  {
          wordPhons.append (phonemeList[y]);
          break;
        }

//    word->setPhonology (wordPhons);
    line.remove (0, line.indexOf ("]") + 2);
    word->setType (line.left (line.indexOf ("...")));
    line.remove (0, line.indexOf ("...") + 3);
    word->setSubType (line.left (line.indexOf (")")));
    line.remove (0, line.indexOf (")") + 1);
    word->setDefinition (line.replace (";;", "\n"));
    wordList.append (word);

    line = in.readLine ();
  }

  phonologyPage->setList (phonemeList);
  phonologyPage->setAlphabeticList (phonemeList);
  suprasegmentalsPage->setPhonemeList (phonemeList);
  phonotacticsPage->setPhonemeList (phonemeList);
  wordPage->setPhonemeList (phonemeList);
  wordPage->setAlphabetOrder (phonemeList);
  wordPage->setWordList (wordList);
  searchPage->setWordList (wordList);
}

void Dictionary::loadPartialWordlist (QDomElement wordListElement)  {
  if (loadPartialDialog)
    delete loadPartialDialog;

  loadPartialDialog = new LoadPartialWordlistDialog (wordListElement, wordPage->getPhonotactics ());

  connect (loadPartialDialog, SIGNAL (loadList (QList< QPointer<Word> >, bool)),
           wordPage, SLOT (addWords (QList< QPointer<Word> >, bool)));

  loadPartialDialog->show ();
}

void Dictionary::loadText (QString filename, QString formula, bool overwrite)  {
  // Preconditions (verified by DictionaryManagementPage::loadText ()):
  // Filename is a real, readable, text file.
  // Formula contains no more than one each of "/w" "/t", "/s" and "/d".
  // Interpreting "/w", "/t", "/s", and "/d" as ".+", and escaping all () symbols,
  // every line of the file is either a blank line, or exactly matches the formula.

  int wordLoc = formula.indexOf ("/w");
  int typeLoc = formula.indexOf ("/t");
  int subtypeLoc = formula.indexOf ("/s");
  int definitionLoc = formula.indexOf ("/d");
  putInOrder (&wordLoc, &typeLoc, &subtypeLoc, &definitionLoc);

  formula.replace ("(", "\\(");
  formula.replace (")", "\\)");
  formula.replace ("/w", "(.+)");
  formula.replace ("/t", "(.+)");
  formula.replace ("/s", "(.+)");
  formula.replace ("/d", "(.+)");
  QRegExp regExp (formula);

  QFile file (filename);
  QTextStream in (&file);
  file.open (QIODevice::ReadOnly | QIODevice::Text);
  QString line = "";
  QList< QPointer<Word> > list;
  Phonotactics phonotax = wordPage->getPhonotactics ();

  while (!in.atEnd ())  {
    line = in.readLine ();

    if (line == "") continue;

    if (!regExp.exactMatch (line)) continue;

    QPointer<Word> newWord = new Word (regExp.cap (wordLoc), phonotax);
    newWord->setType (regExp.cap (typeLoc));
    newWord->setSubType (regExp.cap (subtypeLoc));
    newWord->setDefinition (regExp.cap (definitionLoc));

    list.append (newWord);
  };

  wordPage->addWords (list, overwrite);
}

void Dictionary::loadLexique (QString filename, QStringList paradigm, QList<int> use,
                              bool overwrite)  {
  for (int x = 0; x < paradigm.size (); x++)
    paradigm[x].prepend ("\\");

  QFile file (filename);
  QTextStream in (&file);
  QString line = in.readLine ();
  int currentPara = 0;

  file.open (QIODevice::ReadOnly | QIODevice::Text);

  QList< QPointer<Word> > wList;

  while (!in.atEnd ())  {
    currentPara = 0;
    QStringList values;
    for (int x = 0; x < paradigm.size (); x++)
      values.append ("");

    while (currentPara < paradigm.size ())  {
      if (line == "" || line.split (' ').size () == 0)  {
        if (in.atEnd ()) break;
        line = in.readLine ();
        continue;
      }

      QString para = line.split (' ')[0];

      if (para == paradigm[currentPara])  {
        values[currentPara] = line.remove (0, para.size () + 1);
        currentPara++;
      }

      else if (paradigm.contains (para))  {
        if (paradigm.indexOf (para) < currentPara) break;

        currentPara = paradigm.indexOf (para);
        values[currentPara] = line.remove (0, para.size () + 1);
        currentPara++;
      }

      if (in.atEnd ()) break;
      line = in.readLine ();
    };

    QString name = "";
    QString type = "";
    QString subtype = "";
    QString def = "";

    for (int x = 0; x < values.size (); x++)
      if (use.size () > x && values[x] != "")  {
        switch (use[x])  {
          case 0: name = values[x]; break;
          case 1: type = values[x]; break;
          case 2: subtype = values[x]; break;
          case 3: def = values[x]; break;
          default: break;
        }
      }

    if (name != "")  {
      Word *w = new Word (name, wordPage->getPhonotactics ());
      w->setType (type);
      w->setSubType (subtype);
      w->setDefinition (def);
      wList.append (w);
    }
  };

  file.close ();

  wordPage->addWords (wList, overwrite);
}

void Dictionary::save (QString filename)  {
  QList< QPointer<Phoneme> > phonemeList = phonologyPage->getList ();
  QList< QPointer<Word> > wordList = wordPage->getList ();
  QList< QPointer<Suprasegmental> > supraList = suprasegmentalsPage->getSupraList ();
  Phonotactics phonotactics = wordPage->getPhonotactics ();

  QDomDocument doc ("ConlangML");
  QDomElement root = doc.createElement ("conlang");
  root.setAttribute ("name", languageName);
  doc.appendChild (root);

  if (phonologyPage->useSquareBrackets ())  {
    QDomElement sbTag = doc.createElement ("squareBrackets");
    root.appendChild (sbTag);
  }

  QDomElement inventoryTag = doc.createElement ("inventory");
  root.appendChild (inventoryTag);

  for (int x = 0; x < phonemeList.size (); x++)  {
    QDomElement phonemeTag = doc.createElement ("phonemeDef");
    phonemeTag.setAttribute ("name", phonemeList[x]->getName ());
    inventoryTag.appendChild (phonemeTag);

    QDomElement spellingTag = doc.createElement ("orthography");
    phonemeTag.appendChild (spellingTag);
    QDomText spellingsText = doc.createTextNode (phonemeList[x]->getSpellingsText ());
    spellingTag.appendChild (spellingsText);

    QDomElement notesTag = doc.createElement ("notes");
    phonemeTag.appendChild (notesTag);
    QDomText notesText = doc.createTextNode (phonemeList[x]->getNotes ());
    notesTag.appendChild (notesText);
  }

  QList< QPointer<Phoneme> > alphabetOrder = phonologyPage->getAlphabeticList ();
  QDomElement alphabetTag = doc.createElement ("alphabet");
  inventoryTag.appendChild (alphabetTag);

  QString alphText = "";
  for (int x = 0; x < alphabetOrder.size (); x++)
    alphText += alphabetOrder[x]->getName () + " ";
  if (alphText != "") alphText.chop (1);

  QDomText alphabetText = doc.createTextNode (alphText);
  alphabetTag.appendChild (alphabetText);

  QDomElement phonotaxTag = doc.createElement ("phonotactics");
  phonotaxTag.setAttribute ("onsetReq", phonotactics.onsetIsRequired ());
  root.appendChild (phonotaxTag);

  QDomElement onsetTag = doc.createElement ("onsets");
  phonotaxTag.appendChild (onsetTag);
  QDomElement peakTag = doc.createElement ("peaks");
  phonotaxTag.appendChild (peakTag);
  QDomElement codaTag = doc.createElement ("codas");
  phonotaxTag.appendChild (codaTag);
  QDomElement ignoreTag = doc.createElement ("ignore");
  phonotaxTag.appendChild (ignoreTag);
  QDomText ignoreText = doc.createTextNode (phonotactics.getIgnoreText ());
  ignoreTag.appendChild (ignoreText);

  QList< QList< QPointer<Phoneme> > > onsetList = phonotactics.getOnsetList ();
  for (int x = 0; x < onsetList.size (); x++)  {
    QDomElement sequenceTag = doc.createElement ("sequence");
    onsetTag.appendChild (sequenceTag);

    QString sequence = "";
    for (int y = 0; y < onsetList[x].size (); y++)
      sequence += onsetList[x][y]->getName () + " ";
    if (sequence != "") sequence.chop (1);

    QDomText sequenceText = doc.createTextNode (sequence);
    sequenceTag.appendChild (sequenceText);
  }

  QList< QList< QPointer<Phoneme> > > peakList = phonotactics.getPeakList ();
  for (int x = 0; x < peakList.size (); x++)  {
    QDomElement sequenceTag = doc.createElement ("sequence");
    peakTag.appendChild (sequenceTag);

    QString sequence = "";
    for (int y = 0; y < peakList[x].size (); y++)
      sequence += peakList[x][y]->getName () + " ";
    if (sequence != "") sequence.chop (1);

    QDomText sequenceText = doc.createTextNode (sequence);
    sequenceTag.appendChild (sequenceText);
  }

  QList< QList< QPointer<Phoneme> > > codaList = phonotactics.getCodaList ();
  for (int x = 0; x < codaList.size (); x++)  {
    QDomElement sequenceTag = doc.createElement ("sequence");
    codaTag.appendChild (sequenceTag);

    QString sequence = "";
    for (int y = 0; y < codaList[x].size (); y++)
      sequence += codaList[x][y]->getName () + " ";
    if (sequence != "") sequence.chop (1);

    QDomText sequenceText = doc.createTextNode (sequence);
    sequenceTag.appendChild (sequenceText);
  }

  QDomElement suprasTag = doc.createElement ("suprasegmentals");
  root.appendChild (suprasTag);

  for (int x = 0; x < supraList.size (); x++)  {
    QDomElement supraDef = doc.createElement ("supraDef");
    supraDef.setAttribute ("name", supraList[x]->getName ());
    supraDef.setAttribute ("domain", supraList[x]->getDomain ());
    suprasTag.appendChild (supraDef);

    QDomElement appPhons = doc.createElement ("applies");
    supraDef.appendChild (appPhons);
    QDomText phonText = doc.createTextNode (supraList[x]->getPhonemesText ());
    appPhons.appendChild (phonText);

    QDomElement spelling = doc.createElement ("spelling");
    spelling.setAttribute ("type", supraList[x]->getSpellingType ());
    supraDef.appendChild (spelling);
    QDomText spellingDetail = doc.createTextNode (supraList[x]->getSpellingDetail ());
    spelling.appendChild (spellingDetail);

    QDomElement representation = doc.createElement ("representation");
    representation.setAttribute ("type", supraList[x]->getRepresentationType ());
    supraDef.appendChild (representation);
    QDomText repDetail = doc.createTextNode (supraList[x]->getRepresentationDetail ());
    representation.appendChild (repDetail);

    QDomElement notes = doc.createElement ("notes");
    supraDef.appendChild (notes);
    QDomText notesText = doc.createTextNode (supraList[x]->getDescription ());
    notes.appendChild (notesText);
  }

  QDomElement wordTag = doc.createElement ("wordList");
  root.appendChild (wordTag);

  for (int x = 0; x < wordList.size (); x++)  {
    QDomElement wordDef = doc.createElement ("wordDef");
    wordDef.setAttribute ("name", wordList[x]->getName ());
    wordDef.setAttribute ("type", wordList[x]->getType ());
    wordDef.setAttribute ("subtype", wordList[x]->getSubType ());
    wordTag.appendChild (wordDef);

    QDomElement phonology = doc.createElement ("phonology");
    wordDef.appendChild (phonology);

    QList<Syllable> syllList = wordList[x]->getSyllableList ();
    for (int l = 0; l < syllList.size (); l++)  {
      QDomElement syllable = doc.createElement ("syllable");
      if (syllList[l].getSupraByRep (BEFORE))
        syllable.setAttribute ("bSupra", syllList[l].getSupraByRep (BEFORE)->getName ());
      if (syllList[l].getSupraByRep (FOLLOWING))
        syllable.setAttribute ("fSupra", syllList[l].getSupraByRep (FOLLOWING)->getName ());
      if (syllList[l].getSupraByRep (DOUBLED))
        syllable.setAttribute ("pSupra", syllList[l].getSupraByRep (DOUBLED)->getName ());
      phonology.appendChild (syllable);

      QDomElement onsetTag = doc.createElement ("onset");
      syllable.appendChild (onsetTag);

      QList< QPointer<Phoneme> > onset = syllList[l].getOnset ();
      QList< QList< QPointer<Suprasegmental> > > onsetSupras = syllList[l].getOnsetSupras ();
      for (int o = 0; o < onset.size (); o++)  {
        QDomElement phoneme = doc.createElement ("phoneme");
        phoneme.setAttribute ("name", onset[o]->getName ());
        onsetTag.appendChild (phoneme);

        for (int s = 0; s < onsetSupras[o].size (); s++)  {
          QDomElement supra = doc.createElement ("supra");
          phoneme.appendChild (supra);
          supra.appendChild (doc.createTextNode (onsetSupras[o][s]->getName ()));
        }
      }

      QDomElement peakTag = doc.createElement ("peak");
      syllable.appendChild (peakTag);

      QList< QPointer<Phoneme> > peak = syllList[l].getPeak ();
      QList< QList< QPointer<Suprasegmental> > > peakSupras = syllList[l].getPeakSupras ();
      for (int p = 0; p < peak.size (); p++)  {
        QDomElement phoneme = doc.createElement ("phoneme");
        phoneme.setAttribute ("name", peak[p]->getName ());
        peakTag.appendChild (phoneme);

        for (int s = 0; s < peakSupras[p].size (); s++)  {
          QDomElement supra = doc.createElement ("supra");
          phoneme.appendChild (supra);
          supra.appendChild (doc.createTextNode (peakSupras[p][s]->getName ()));
        }
      }

      QDomElement codaTag = doc.createElement ("coda");
      syllable.appendChild (codaTag);

      QList< QPointer<Phoneme> > coda = syllList[l].getCoda ();
      QList< QList< QPointer<Suprasegmental> > > codaSupras = syllList[l].getCodaSupras ();
      for (int c = 0; c < coda.size (); c++)  {
        QDomElement phoneme = doc.createElement ("phoneme");
        phoneme.setAttribute ("name", coda[c]->getName ());
        codaTag.appendChild (phoneme);

        for (int s = 0; s < codaSupras[c].size (); s++)  {
          QDomElement supra = doc.createElement ("supra");
          phoneme.appendChild (supra);
          supra.appendChild (doc.createTextNode (codaSupras[c][s]->getName ()));
        }
      }
    }

    QDomElement def = doc.createElement ("definition");
    wordDef.appendChild (def);
    def.appendChild (doc.createTextNode (wordList[x]->getDefinition ()));
  }

  QFile file (filename + ".cdic");

  if (!file.open (QFile::WriteOnly | QFile::Text))  {
    QMessageBox::warning (this, "", "Cannot write file.");
    return;
  }

  file.write (doc.toByteArray ());

  file.close ();

/*
  for (int x = 0; x < phonemeList.size (); x++)  {
    QString newName = phonemeList[x]->getName ().remove (" ");
    newName = newName.remove ("]");
    phonemeList[x]->setName (newName);

    QString newSpellings = phonemeList[x]->getSpellingsText ().replace (QRegExp ("\\s+"), " ");
    phonemeList[x]->setSpellings (newSpellings);

    QString newDesc = phonemeList[x]->getNotes ().replace (QRegExp ("\r?\n"), ";;");
    phonemeList[x]->setNotes (newDesc);

    out << phonemeList[x]->getName () << "  "
        << phonemeList[x]->getSpellingsText () << "  "
        << phonemeList[x]->getNotes () << "\r\n";
  }

  out << "\r\n";

  for (int x = 0; x < wordList.size (); x++)  {
    QString newName = wordList[x]->getName ().remove ("[");
    QString newType = wordList[x]->getType ().replace ("...", "..");
    QString newSubType = wordList[x]->getSubType ().remove (")");
    QString newDef = wordList[x]->getDefinition ().replace
                                                    (QRegExp ("\r?\n"), ";;");

    out << newName << "[";

    QList< QPointer<Phoneme> > wordPhons = wordList[x]->getPhonemeList ();
    for (int y = 0; y < wordPhons.size (); y++)  {
      out << wordPhons[y]->getName ();
      if (y != wordPhons.size () - 1) out << " ";
    }
    out << "](" << newType << "..." << newSubType << ")" << newDef << "\r\n";
  }
  */
}

void Dictionary::saveText (QString filename, QString formula, bool overwrite)  {
  QFile file (filename);
  QTextStream out (&file);

  if (overwrite)
    file.open (QIODevice::WriteOnly | QIODevice::Text);
  else file.open (QIODevice::WriteOnly | QIODevice::Text | QIODevice::Append);

  QList< QPointer<Word> > wordList = wordPage->getList ();

  for (int x = 0; x < wordList.size (); x++)  {
    QString line = formula;
    line.replace ("/w", wordList[x]->getName ());
    line.replace ("/t", wordList[x]->getType ());
    line.replace ("/s", wordList[x]->getSubType ());
    line.replace ("/p", wordList[x]->getPhonology ());
    line.replace ("/d", wordList[x]->getDefinition ());

    out << line << endl;
  }

  file.close ();
}

void Dictionary::setLanguageName (QString name)  {
  languageName = name;
  emit languageNameUpdated (name);
}

void Dictionary::putInOrder (int *first, int *second, int *third, int *fourth)  {
  // Takes four int pointers, and assigns all non-negative values the numbers
  // one through four, based on the order of the numbers from lowest to highest.

  int *array[4] = {first, second, third, fourth};

  for (int x = 0; x < 3; x++)  {
    int smallest = x;

    for (int y = x+1; y < 4; y++)
      if (*array[y] < *array[smallest])
        smallest = y;

    int *temp = array[x];
    array[x] = array[smallest];
    array[smallest] = temp;
  }

  int number = 1;

  for (int x = 0; x < 4; x++)
    if (*array[x] >= 0)
      *array[x] = number++;
}