#include <QString>
#include <QStringList>
#include <QListView>
#include <QStringListModel>
#include <QLabel>
#include <QPushButton>
#include <QCheckBox>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QDomDocument>
#include <QDomNodeList>
#include <QItemSelectionModel>
#include <QModelIndex>
#include <QModelIndexList>
#include <QMessageBox>
#include <QRegExp>

#include "loadpartialwordlistdialog.h"
#include "word.h"

LoadPartialWordlistDialog::LoadPartialWordlistDialog (QDomElement wordElement, Phonotactics phonotax) {
  phonotactics = phonotax;
  wordList = wordElement;

  includedWords = QStringList ();
  excludedWords = QStringList ();

  QDomNodeList wordNodes = wordList.childNodes ();
  for (int x = 0; x < wordNodes.size (); x++)  {
    QDomElement currentWord = wordNodes.at (x).toElement ();
    if (currentWord.isNull ()) continue;

    if (currentWord.tagName () == "wordDef")  {
      QString wordEntry = currentWord.attribute ("name");

      if (currentWord.attribute ("type") != "")  {
        wordEntry += " (" + currentWord.attribute ("type");

        if (currentWord.attribute ("subtype") != "")
          wordEntry += ", " + currentWord.attribute ("subtype");

        wordEntry += ")";
      }

      excludedWords.append (wordEntry);
    }
  }

  sortLists ();

  excludedWordsLabel = new QLabel ("Excluded Words");
  excludedWordsLabel->setMaximumSize (excludedWordsLabel->sizeHint ());
  excludedWordsModel = new QStringListModel (excludedWords, this);
  excludedWordsView = new QListView;
  excludedWordsView->setModel (excludedWordsModel);
  excludedWordsView->setSelectionBehavior (QAbstractItemView::SelectRows);
  excludedWordsView->setSelectionMode (QAbstractItemView::ExtendedSelection);
  excludedWordsView->setEditTriggers (QAbstractItemView::NoEditTriggers);

  excludedWordsLayout = new QVBoxLayout;
  excludedWordsLayout->addWidget (excludedWordsLabel);
  excludedWordsLayout->setAlignment (excludedWordsLabel, Qt::AlignCenter);
  excludedWordsLayout->addWidget (excludedWordsView);

  includedWordsLabel = new QLabel ("Included Words");
  includedWordsLabel->setMaximumSize (includedWordsLabel->sizeHint ());
  includedWordsModel = new QStringListModel (QStringList (), this);
  includedWordsView = new QListView;
  includedWordsView->setModel (includedWordsModel);
  includedWordsView->setSelectionBehavior (QAbstractItemView::SelectRows);
  includedWordsView->setSelectionMode (QAbstractItemView::ExtendedSelection);
  includedWordsView->setEditTriggers (QAbstractItemView::NoEditTriggers);

  includedWordsLayout = new QVBoxLayout;
  includedWordsLayout->addWidget (includedWordsLabel);
  includedWordsLayout->setAlignment (includedWordsLabel, Qt::AlignCenter);
  includedWordsLayout->addWidget (includedWordsView);

  addButton = new QPushButton ("Add ->");
  addButton->setMaximumSize (addButton->sizeHint ());
  addButton->setEnabled (false);
  removeButton = new QPushButton ("<- Remove");
  removeButton->setMaximumSize (removeButton->sizeHint ());
  removeButton->setEnabled (false);

  addRemoveLayout = new QVBoxLayout;
  addRemoveLayout->addStretch (2);
  addRemoveLayout->addWidget (addButton);
  addRemoveLayout->setAlignment (addButton, Qt::AlignCenter);
  addRemoveLayout->addStretch (1);
  addRemoveLayout->addWidget (removeButton);
  addRemoveLayout->addStretch (2);

  listViewsLayout = new QHBoxLayout;
  listViewsLayout->addLayout (excludedWordsLayout);
  listViewsLayout->addLayout (addRemoveLayout);
  listViewsLayout->addLayout (includedWordsLayout);

  overwriteBox = new QCheckBox ("Overwrite identical entries");
  overwriteBox->setMaximumSize (overwriteBox->sizeHint ());
  importButton = new QPushButton ("Import");
  importButton->setMaximumSize (importButton->sizeHint ());

  mainLayout = new QVBoxLayout;
  mainLayout->addLayout (listViewsLayout);
  mainLayout->addWidget (overwriteBox);
  mainLayout->setAlignment (overwriteBox, Qt::AlignCenter);
  mainLayout->addWidget (importButton);
  mainLayout->setAlignment (importButton, Qt::AlignCenter);

  setLayout (mainLayout);

  connect (excludedWordsView->selectionModel (),
           SIGNAL (selectionChanged (const QItemSelection, const QItemSelection)), this,
           SLOT (excludedSelected ()));
  connect (excludedWordsView->selectionModel (),
           SIGNAL (selectionChanged (const QItemSelection, const QItemSelection)),
           includedWordsView->selectionModel (), SLOT (reset ()));
  connect (includedWordsView->selectionModel (),
           SIGNAL (selectionChanged (const QItemSelection, const QItemSelection)), this,
           SLOT (includedSelected ()));
  connect (includedWordsView->selectionModel (),
           SIGNAL (selectionChanged (const QItemSelection, const QItemSelection)),
           excludedWordsView->selectionModel (), SLOT (reset ()));
  connect (addButton, SIGNAL (clicked ()), this, SLOT (includeWords ()));
  connect (removeButton, SIGNAL (clicked ()), this, SLOT (excludeWords ()));
  connect (importButton, SIGNAL (clicked ()), this, SLOT (import ()));
  connect (importButton, SIGNAL (clicked ()), this, SLOT (hide ()));
}

void LoadPartialWordlistDialog::excludedSelected ()  {
  addButton->setEnabled (true);
  removeButton->setEnabled (false);
}

void LoadPartialWordlistDialog::includedSelected ()  {
  addButton->setEnabled (false);
  removeButton->setEnabled (true);
}

void LoadPartialWordlistDialog::includeWords ()  {
  QModelIndexList indexList = excludedWordsView->selectionModel ()->selectedRows ();
  QList<int> rows;

  for (int x = 0; x < indexList.size (); x++)
    rows.append (indexList[x].row ());

  qSort (rows.begin (), rows.end ());

  for (int y = rows.size () - 1; y >= 0; y--)  {
    includedWords.append (excludedWords[rows[y]]);
    excludedWords.removeAt (rows[y]);
  }

  sortLists ();
  updateModels ();
}

void LoadPartialWordlistDialog::excludeWords ()  {
  QModelIndexList indexList = includedWordsView->selectionModel ()->selectedRows ();
  QList<int> rows;

  for (int x = 0; x < indexList.size (); x++)
    rows.append (indexList[x].row ());

  qSort (rows.begin (), rows.end ());

  for (int y = rows.size () - 1; y >= 0; y--)  {
    excludedWords.append (includedWords[rows[y]]);
    includedWords.removeAt (rows[y]);
  }

  sortLists ();
  updateModels ();
}

void LoadPartialWordlistDialog::import ()  {
  if (includedWords.isEmpty ())  {
    QMessageBox::warning (this, "", "You haven't selected any words to import!");
    return;
  }

  QList< QPointer<Word> > importList;
  QRegExp regExp ("(\\w+) (?:\\((.+)(?:, (.+))?\\))?");

  for (int x = 0; x < includedWords.size (); x++)  {
    if (regExp.exactMatch (includedWords[x]))  {
      QPointer<Word> word = new Word (regExp.cap (1), phonotactics);
      word->setType (regExp.cap (2));
      word->setSubType (regExp.cap (3));

      QDomNodeList nodeList = wordList.childNodes ();
      for (int w = 0; w < nodeList.size (); w++)  {
        QDomElement element = nodeList.at (w).toElement ();

        if (element.attribute ("name") == regExp.cap (1) &&
            element.attribute ("type") == regExp.cap (2) &&
            element.attribute ("subtype") == regExp.cap (3))  {
          word->setDefinition (element.firstChildElement ("definition").text ());
          break;
        }
      }

      importList.append (word);
    }
  }

  emit loadList (importList, overwriteBox->isChecked ());
}

void LoadPartialWordlistDialog::sortLists ()  {
  for (int x = 0; x < includedWords.size (); x++)  {
    int smallest = x;

    for (int y = x; y < includedWords.size (); y++)
      if (includedWords[y].localeAwareCompare (includedWords[smallest]) < 0)
        smallest = y;

    if (smallest != x) includedWords.swap (x, smallest);
  }

  for (int x = 0; x < excludedWords.size (); x++)  {
    int smallest = x;

    for (int y = x; y < excludedWords.size (); y++)
      if (excludedWords[y].localeAwareCompare (excludedWords[smallest]) < 0)
        smallest = y;

    if (smallest != x) excludedWords.swap (x, smallest);
  }
}

void LoadPartialWordlistDialog::updateModels ()  {
  excludedWordsView->selectionModel ()->reset ();
  includedWordsView->selectionModel ()->reset ();
  addButton->setEnabled (false);
  removeButton->setEnabled (false);

  excludedWordsModel->setStringList (excludedWords);
  includedWordsModel->setStringList (includedWords);
}