#include <QStringListModel>
#include <QPushButton>
#include <QListView>
#include <QRadioButton>
#include <QLineEdit>
#include <QLabel>
#include <QTextEdit>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QTextStream>

#include <iostream>
using namespace std;

#include "phonologypage.h"
#include "phoneme.h"

PhonologyPage::PhonologyPage ()  {
  squareBrackets = false;
  phonemeModel = new QStringListModel;

  addPhonemeButton = new QPushButton ("Add Phoneme(s)");
  addPhonemeButton->setMaximumSize (addPhonemeButton->sizeHint ());
  addPhonemeEdit = new QLineEdit ();
  addPhonemeLayout = new QHBoxLayout;
  addPhonemeLayout->addWidget (addPhonemeButton);
  addPhonemeLayout->addWidget (addPhonemeEdit);

  phonemeListView = new QListView;
  phonemeListView->setModel (phonemeModel);
  phonemeListView->setSelectionBehavior (QAbstractItemView::SelectRows);
  phonemeListView->setEditTriggers (QAbstractItemView::NoEditTriggers);
  phonemeListView->setSelectionMode (QAbstractItemView::ExtendedSelection);

  parsingOrderButton = new QRadioButton ("Show Parsing Order");
  parsingOrderButton->setChecked (true);
  alphabetOrderButton = new QRadioButton ("Show Alphabetic Order");

  listLayout = new QVBoxLayout;
  listLayout->addLayout (addPhonemeLayout);
  listLayout->addWidget (phonemeListView);
  listLayout->addWidget (parsingOrderButton);
  listLayout->addWidget (alphabetOrderButton);

  moveUpButton = new QPushButton ("Move Up");
  moveUpButton->setMaximumSize (moveUpButton->sizeHint ());
  moveUpButton->setEnabled (false);
  moveDownButton = new QPushButton ("Move Down");
  moveDownButton->setMaximumSize (moveDownButton->sizeHint ());
  moveDownButton->setEnabled (false);
  deleteButton = new QPushButton ("Delete");
  deleteButton->setMaximumSize (deleteButton->sizeHint ());
  deleteButton->setEnabled (false);
  moveUpDownLayout = new QVBoxLayout;
  moveUpDownLayout->addStretch (2);
  moveUpDownLayout->addWidget (moveUpButton);
  moveUpDownLayout->setAlignment (moveUpButton, Qt::AlignCenter);
  moveUpDownLayout->addStretch (1);
  moveUpDownLayout->addWidget (moveDownButton);
  moveUpDownLayout->addStretch (1);
  moveUpDownLayout->addWidget (deleteButton);
  moveUpDownLayout->setAlignment (deleteButton, Qt::AlignCenter);

  QLabel *phonemeNameLabel = new QLabel ("Phoneme Name:");
  phonemeNameEdit = new QLineEdit ();
  phonemeNameEdit->setEnabled (false);
  phonemeNameButton = new QPushButton ("Change");
  phonemeNameButton->setMaximumSize (phonemeNameButton->sizeHint ());
  phonemeNameButton->setEnabled (false);
  phonemeNameLayout = new QHBoxLayout;
  phonemeNameLayout->addWidget (phonemeNameLabel);
  phonemeNameLayout->addWidget (phonemeNameEdit);
  phonemeNameLayout->addWidget (phonemeNameButton);

  QLabel *phonemeSpellingLabel =
        new QLabel ("Spellings (separated by spaces):");
  phonemeSpellingEdit = new QLineEdit ();
  phonemeSpellingEdit->setEnabled (false);
  phonemeSpellingLayout = new QHBoxLayout;
  phonemeSpellingLayout->addWidget (phonemeSpellingLabel);
  phonemeSpellingLayout->addWidget (phonemeSpellingEdit);

  QLabel *notesLabel = new QLabel ("Notes:");
  notesEdit = new QTextEdit ();
  notesEdit->setEnabled (false);

  phonemeInfoLayout = new QVBoxLayout;
  phonemeInfoLayout->addLayout (phonemeNameLayout);
  phonemeInfoLayout->addLayout (phonemeSpellingLayout);
  phonemeInfoLayout->addWidget (notesLabel);
  phonemeInfoLayout->addWidget (notesEdit);

  mainLayout = new QHBoxLayout;
  mainLayout->addLayout (listLayout);
  mainLayout->addLayout (moveUpDownLayout);
  mainLayout->addLayout (phonemeInfoLayout);

  setLayout (mainLayout);

  connect (addPhonemeButton, SIGNAL (clicked ()), this, SLOT (addPhoneme ()));
  connect (addPhonemeButton, SIGNAL (clicked ()), addPhonemeEdit,
           SLOT (clear ()));
  connect (parsingOrderButton, SIGNAL (toggled (bool)), this,
           SLOT (updateModel ()));
  connect (alphabetOrderButton, SIGNAL (toggled (bool)), this,
           SLOT (updateModel ()));
  connect (moveUpButton, SIGNAL (clicked ()), this, SLOT (moveUp ()));
  connect (moveDownButton, SIGNAL (clicked ()), this, SLOT (moveDown ()));
  connect (deleteButton, SIGNAL (clicked ()), this, SLOT (deletePhoneme ()));
  connect (phonemeNameButton, SIGNAL (clicked ()), this,
           SLOT (editPhonemeName ()));
  connect (phonemeSpellingEdit, SIGNAL (textEdited (QString)), this,
           SLOT (editPhonemeSpelling (QString)));
  connect (notesEdit, SIGNAL (textChanged ()), this,
           SLOT (editPhonemeNotes ()));
}

void PhonologyPage::setModel (QStringListModel *model)  {
  phonemeModel = model;
  phonemeListView->setModel (phonemeModel);

  connect (phonemeListView->selectionModel (),
           SIGNAL (currentChanged (const QModelIndex&, const QModelIndex&)),
           this, SLOT (displayPhoneme ()));
  connect (phonemeListView->selectionModel (),
           SIGNAL (selectionChanged (const QItemSelection&, const QItemSelection&)),
           this, SLOT (updateGUI ()));
}

QList< QPointer<Phoneme> > PhonologyPage::getList ()  {
  return phonemeList;
}

QList< QPointer<Phoneme> > PhonologyPage::getAlphabeticList ()  {
  return alphabetOrder;
}

bool PhonologyPage::useSquareBrackets ()  {
  return squareBrackets;
}

void PhonologyPage::setList (QList< QPointer<Phoneme> > list)  {
  phonemeList = list;
  updateModel ();
}

void PhonologyPage::setAlphabeticList (QList< QPointer<Phoneme> > list)  {
  alphabetOrder = list;
  updateModel ();
}

void PhonologyPage::setPhonemeFont (QFont font)  {
  phonemeListView->setFont (font);
  addPhonemeEdit->setFont (font);
  phonemeNameEdit->setFont (font);
}

void PhonologyPage::setOrthographyFont (QFont font)  {
  phonemeSpellingEdit->setFont (font);
}

void PhonologyPage::setNotesFont (QFont font)  {
  notesEdit->setFont (font);
}

void PhonologyPage::setSquareBrackets (bool sb)  {
  squareBrackets = sb;
  updateModel ();
}

void PhonologyPage::clear ()  {
  phonemeList.clear ();
  alphabetOrder.clear ();
  updateModel ();
}

void PhonologyPage::addPhoneme ()  {
  if (addPhonemeEdit->text () == "") return;

  QStringList newPhonemes = addPhonemeEdit->text ().split (' ');

  for (int x = 0; x < newPhonemes.size (); x++)  {
    phonemeList.append (new Phoneme (newPhonemes[x]));
    alphabetOrder.append (new Phoneme (newPhonemes[x]));
  }

  emit phonemeListUpdated (phonemeList);
  emit alphabetUpdated (alphabetOrder);
  updateModel ();
}

void PhonologyPage::displayPhoneme ()  {
  int phonemeNum = phonemeListView->selectionModel ()->currentIndex ().row ();

  phonemeNameEdit->setEnabled (true);
  phonemeNameButton->setEnabled (true);
  phonemeSpellingEdit->setEnabled (true);
  notesEdit->setEnabled (true);

  updateGUI ();

  if (parsingOrderButton->isChecked ())  {
    phonemeNameEdit->setText (phonemeList[phonemeNum]->getName ());
    phonemeSpellingEdit->setText (phonemeList[phonemeNum]->getSpellingsText ());
    notesEdit->setPlainText (phonemeList[phonemeNum]->getNotes ());
  }

  else  {
    phonemeNameEdit->setText (alphabetOrder[phonemeNum]->getName ());
    phonemeSpellingEdit->setText (alphabetOrder[phonemeNum]->getSpellingsText ());
    notesEdit->setPlainText (alphabetOrder[phonemeNum]->getNotes ());
  }
}

void PhonologyPage::updateGUI ()  {
  QModelIndexList indexList = phonemeListView->selectionModel ()->selectedRows ();

  if (indexList.size () == 0)  {
    moveUpButton->setEnabled (false);
    moveDownButton->setEnabled (false);
    deleteButton->setEnabled (false);
    return;
  }

  moveUpButton->setEnabled (true);
  moveDownButton->setEnabled (true);
  deleteButton->setEnabled (true);

  for (int x = 0; x < indexList.size (); x++)  {
    if (indexList[x].row () == 0)
      moveUpButton->setEnabled (false);
    if (indexList[x].row () == (phonemeList.size () - 1))
      moveDownButton->setEnabled (false);
  }
}

void PhonologyPage::moveUp ()  {
  QModelIndexList indexList = phonemeListView->selectionModel ()->selectedRows ();
  QList<int> rowList;
  for (int x = 0; x < indexList.size (); x++)
    rowList.append (indexList[x].row ());

  if (parsingOrderButton->isChecked ()) {
    for (int x = 0; x < rowList.size (); x++)
      phonemeList.swap (rowList[x], rowList[x] - 1);
    emit phonemeListUpdated (phonemeList);
  }

  else  {
    for (int x = 0; x < rowList.size (); x++)
      alphabetOrder.swap (rowList[x], rowList[x] - 1);
    emit alphabetUpdated (alphabetOrder);
  }

  updateModel ();

  for (int x = 0; x < rowList.size (); x++)
    phonemeListView->selectionModel ()->select
      (phonemeModel->index (rowList[x] - 1), QItemSelectionModel::Toggle);
}

void PhonologyPage::moveDown ()  {
  QModelIndexList indexList = phonemeListView->selectionModel ()->selectedRows ();
  QList<int> rowList;
  for (int x = 0; x < indexList.size (); x++)
    rowList.append (indexList[x].row ());

  if (parsingOrderButton->isChecked ()) {
    for (int x = 0; x < rowList.size (); x++)
      phonemeList.swap (rowList[x], rowList[x] + 1);
    emit phonemeListUpdated (phonemeList);
  }

  else  {
    for (int x = 0; x < rowList.size (); x++)
      alphabetOrder.swap (rowList[x], rowList[x] + 1);
    emit alphabetUpdated (alphabetOrder);
  }

  updateModel ();

  for (int x = 0; x < rowList.size (); x++)
    phonemeListView->selectionModel ()->select
      (phonemeModel->index (rowList[x] + 1), QItemSelectionModel::Toggle);
}

void PhonologyPage::deletePhoneme ()  {
  QModelIndexList rowList = phonemeListView->selectionModel ()->selectedRows ();

  for (int x = 0; x < rowList.size () - 1; x++)  {
    int largest = x;

    for (int y = x + 1; y < rowList.size (); y++)
      if (rowList[y].row () > rowList[largest].row ())
        largest = y;

    if (largest != x)
      rowList.swap (largest, x);
  }

  for (int x = 0; x < rowList.size (); x++)  {
    QPointer<Phoneme> p;

    if (parsingOrderButton->isChecked ())  {
      p = phonemeList[rowList[x].row ()];
      alphabetOrder.removeAll (phonemeList[rowList[x].row ()]);
      phonemeList.removeAt (rowList[x].row ());
    }

    else  {
      p = alphabetOrder[rowList[x].row ()];
      phonemeList.removeAll (alphabetOrder[rowList[x].row ()]);
      alphabetOrder.removeAt (rowList[x].row ());
    }
  
    delete p;
  }
  
  emit phonemeListUpdated (phonemeList);
  emit alphabetUpdated (alphabetOrder);
  updateModel ();
}

void PhonologyPage::editPhonemeName ()  {
  QModelIndex index = phonemeListView->selectionModel ()->currentIndex ();
  QString name = phonemeNameEdit->text ();

  if (name == "") return;

  if (parsingOrderButton->isChecked ())
    phonemeList[index.row ()]->setName (name);
  else alphabetOrder[index.row ()]->setName (name);

  updateModel ();
  phonemeListView->selectionModel ()->setCurrentIndex
                                  (index, QItemSelectionModel::SelectCurrent);
}

void PhonologyPage::editPhonemeSpelling (QString spelling)  {
  int phonemeNum = phonemeListView->selectionModel ()->currentIndex ().row ();

  if (parsingOrderButton->isChecked ())
    phonemeList[phonemeNum]->setSpellings (spelling);
  else alphabetOrder[phonemeNum]->setSpellings (spelling);
}

void PhonologyPage::editPhonemeNotes ()  {
  int phonemeNum = phonemeListView->selectionModel ()->currentIndex ().row ();
  if (phonemeNum >= 0 && phonemeNum < phonemeList.size ()) {
    if (parsingOrderButton->isChecked ())
      phonemeList[phonemeNum]->setNotes (notesEdit->toPlainText ());
    else alphabetOrder[phonemeNum]->setNotes (notesEdit->toPlainText ());
  }
}

void PhonologyPage::updateModel ()  {
  phonemeListView->selectionModel ()->reset ();
  moveUpButton->setEnabled (false);
  moveDownButton->setEnabled (false);
  deleteButton->setEnabled (false);
  phonemeNameEdit->setText ("");
  phonemeNameEdit->setEnabled (false);
  phonemeNameButton->setEnabled (false);
  phonemeSpellingEdit->setText ("");
  phonemeSpellingEdit->setEnabled (false);
  notesEdit->setPlainText ("");
  notesEdit->setEnabled (false);

  QStringList nameList;

  if (parsingOrderButton->isChecked ())
    for (int x = 0; x < phonemeList.size (); x++)  {
      if (squareBrackets) nameList.append ("[" + phonemeList[x]->getName () + "]");
      else nameList.append ("/" + phonemeList[x]->getName () + "/");
    }

  else for (int x = 0; x < alphabetOrder.size (); x++)  {
    if (squareBrackets) nameList.append ("[" + alphabetOrder[x]->getName () + "]");
    else nameList.append ("/" + alphabetOrder[x]->getName () + "/");
  }

  phonemeModel->setStringList (nameList);
}