#include <iostream>
using namespace std;

#include <QString>
#include <QStringList>
#include <QMessageBox>

#include "syllable.h"
#include "phoneme.h"
#include "const.h"

Syllable::Syllable ()  {
  beforeSupra = NULL;
  peakSupra = NULL;
  followingSupra = NULL;
}

QList< QPointer<Phoneme> > Syllable::getOnset ()  {
  return onset;
}

QList< QPointer<Phoneme> > Syllable::getPeak ()  {
  return peak;
}

QList< QPointer<Phoneme> > Syllable::getCoda ()  {
  return coda;
}

QList< QList< QPointer<Suprasegmental> > > Syllable::getOnsetSupras ()  {
  return onsetSupras;
}

QList< QList< QPointer<Suprasegmental> > > Syllable::getPeakSupras ()  {
  return peakSupras;
}

QList< QList< QPointer<Suprasegmental> > > Syllable::getCodaSupras ()  {
  return codaSupras;
}

QStringList Syllable::getOnsetSpellings ()  {
  return onsetSpellings;
}

QStringList Syllable::getPeakSpellings ()  {
  return peakSpellings;
}

QStringList Syllable::getCodaSpellings ()  {
  return codaSpellings;
}

QList< QPointer<Suprasegmental> > Syllable::getSupras ()  {
  return QList< QPointer<Suprasegmental> > () << beforeSupra << followingSupra
                                              << peakSupra;
}

QPointer<Suprasegmental> Syllable::getSupraBySpelling (int spelling)  {
  if (beforeSupra)
    if (beforeSupra->getSpellingType () == spelling)
      return beforeSupra;

  if (followingSupra)
    if (followingSupra->getSpellingType () == spelling)
      return followingSupra;

  if (peakSupra)
    if (peakSupra->getSpellingType () == spelling)
      return peakSupra;

  return NULL;
}

QPointer<Suprasegmental> Syllable::getSupraByRep (int representation)  {
  if (representation == BEFORE)
    return beforeSupra;
  else if (representation == FOLLOWING)
    return followingSupra;
  else return peakSupra;
}

bool Syllable::hasSupra (QPointer<Suprasegmental> supra)  {
  if (peakSupra == supra) return true;
  if (beforeSupra == supra) return true;
  if (followingSupra == supra) return true;
  return false;
}

bool Syllable::onsetHasSupra (int num, QPointer<Suprasegmental> supra)  {
  if (onsetSupras.size () <= num) return false;
  return onsetSupras[num].contains (supra);
}

bool Syllable::peakHasSupra (int num, QPointer<Suprasegmental> supra)  {
  if (peakSupras.size () <= num) return false;
  return peakSupras[num].contains (supra);
}

bool Syllable::codaHasSupra (int num, QPointer<Suprasegmental> supra)  {
  if (codaSupras.size () <= num) return false;
  return codaSupras[num].contains (supra);
}

QString Syllable::getText ()  {
  QString text = "";

  if (beforeSupra)
    text += beforeSupra->getRepresentationDetail ();

  for (int x = 0; x < onset.size (); x++)
    text += applySupras (onset[x], onsetSupras[x]);

  QString peakText = "";

  for (int x = 0; x < peak.size (); x++)
    peakText += applySupras (peak[x], peakSupras[x]);

  if (peakSupra)  {
    if (peakSupra->getRepresentationType () == DIACRITIC)
      peakText = addDiacritic
                  (peakText, peakSupra->getRepresentationDetail ());

    else peakText += peakText;
  }

  text += peakText;

  for (int x = 0; x < coda.size (); x++)
    text += applySupras (coda[x], codaSupras[x]);

  if (followingSupra)
    text += followingSupra->getRepresentationDetail ();

  if (text == "")
    return "?";

  return text;
}

QString Syllable::getSaveOutput ()  {
  QString text = "[";

  for (int x = 0; x < onset.size (); x++)  {
    text += onset[x]->getName ().remove (" ").remove ("]") + " (";

    for (int y = 0; y < onsetSupras[x].size (); y++)
      text += "\"" + onsetSupras[x][y]->getName ().replace ("\"", "\'") + "\",";

    if (text.endsWith (",")) text.chop (1);

    text += ") ";
  }

  text += ". ";

  for (int x = 0; x < peak.size (); x++)  {
    text += peak[x]->getName ().remove (" ").remove ("]") + " (";

    for (int y = 0; y < peakSupras[x].size (); y++)
      text += "\"" + peakSupras[x][y]->getName ().replace ("\"", "\'") + "\",";

    if (text.endsWith (",")) text.chop (1);

    text += ") ";
  }

  text += ". ";

  for (int x = 0; x < coda.size (); x++)  {
    text += coda[x]->getName ().remove (" ").remove ("]") + " (";

    for (int y = 0; y < codaSupras[x].size (); y++)
      text += "\"" + codaSupras[x][y]->getName ().replace ("\"", "\'") + "\",";

    if (text.endsWith (",")) text.chop (1);

    text += ") ";
  }

  if (text.endsWith (" ")) text.chop (1);
  text += "],\"" + beforeSupra->getName ().replace ("\"", "\\\"") + "\",";
  text += "\"" + peakSupra->getName ().replace ("\"", "\\\"") + "\",";
  text += "\"" + followingSupra->getName ().replace ("\"", "\\\"") + "\"";

  return text;
}

QList< QPointer<Suprasegmental> > Syllable::popFirst (QPointer<Phoneme> &p)  {
  QList< QPointer<Suprasegmental> > supras;
  
  if (onset.size () > 0)  {
    p = onset.takeFirst ();
    supras = onsetSupras.takeFirst ();
  }
  else if (peak.size () > 0)  {
    p = peak.takeFirst ();
    supras = peakSupras.takeFirst ();
  }
  else if (coda.size () > 0)  {
    p = coda.takeFirst ();
    supras = codaSupras.takeFirst ();
  }
  else p = NULL;

  return supras;
}

QList< QPointer<Suprasegmental> > Syllable::popLast (QPointer<Phoneme> &p)  {
  QList< QPointer<Suprasegmental> > supras;

  if (coda.size () > 0)  {
    p = coda.takeLast ();
    supras = codaSupras.takeLast ();
  }
  else if (peak.size () > 0)  {
    p = peak.takeLast ();
    supras = peakSupras.takeLast ();
  }
  else if (onset.size () > 0)  {
    p = onset.takeLast ();
    supras = onsetSupras.takeLast ();
  }
  else p = NULL;

  return supras;
}

void Syllable::appendToOnset (QPointer<Phoneme> p, QList< QPointer<Suprasegmental> > s)  {
  if (p == NULL) return;
  
  onset.prepend (p);
  onsetSupras.prepend (s);
}

void Syllable::appendToCoda (QPointer<Phoneme> p, QList< QPointer<Suprasegmental> > s)  {
  if (p == NULL) return;
  
  coda.append (p);
  codaSupras.append (s);
}

void Syllable::setOnset (QList< QPointer<Phoneme> > p,
                         QList< QList< QPointer<Suprasegmental> > > s)  {
  onset = p;
  onsetSupras = s;

  for (int x = 0; x < onset.size (); x++)
    if (onsetSupras.size () <= x)
      onsetSupras.append (QList< QPointer<Suprasegmental> > ());
}

void Syllable::setPeak (QList< QPointer<Phoneme> > p,
                        QList< QList< QPointer<Suprasegmental> > > s)  {
  peak = p;
  peakSupras = s;

  for (int x = 0; x < peakSupras.size (); x++)
    for (int y = 0; y < peakSupras[x].size (); y++)
      if (peakSupras[x][y]->getDomain () == SYLL_APPL)  {
//        QMessageBox::information (NULL, "setPeak",
//                                  "Applying supra to syllable instead of peak: "
//                                  + peakSupras[x][y]->getName ());

        if (peakSupras[x][y]->getRepresentationType () == BEFORE)
          beforeSupra = peakSupras[x][y];
        else if (peakSupras[x][y]->getRepresentationType () == FOLLOWING)
          followingSupra = peakSupras[x][y];
        else peakSupra = peakSupras[x][y];

        peakSupras[x].removeAll (peakSupras[x][y]);
      }

  for (int x = 0; x < peak.size (); x++)
    if (peakSupras.size () <= x)
      peakSupras.append (QList< QPointer<Suprasegmental> > ());
}

void Syllable::setCoda (QList< QPointer<Phoneme> > p,
                        QList< QList< QPointer<Suprasegmental> > > s)  {
  coda = p;
  codaSupras = s;

  for (int x = 0; x < coda.size (); x++)
    if (codaSupras.size () <= x)
      codaSupras.append (QList< QPointer<Suprasegmental> > ());
}

void Syllable::setOnsetSpellings (QStringList os)  {
  onsetSpellings = os;
}

void Syllable::setPeakSpellings (QStringList ps)  {
  peakSpellings = ps;
}

void Syllable::setCodaSpellings (QStringList cs)  {
  codaSpellings = cs;
}

void Syllable::clearSpellings ()  {
  onsetSpellings.clear ();
  peakSpellings.clear ();
  codaSpellings.clear ();
}

void Syllable::addSupra (QPointer<Suprasegmental> s)  {
  if (s->getRepresentationType () == BEFORE)
    beforeSupra = s;
  else if (s->getRepresentationType () == FOLLOWING)
    followingSupra = s;
  else peakSupra = s;
}

void Syllable::removeSupra (QPointer<Suprasegmental> s)  {
  if (s == beforeSupra) beforeSupra = NULL;
  else if (s == followingSupra) followingSupra = NULL;
  else if (s == peakSupra) peakSupra = NULL;
}

void Syllable::clearSupraOfType (int type)  {
  if (type == BEFORE) beforeSupra = NULL;
  else if (type == FOLLOWING) followingSupra = NULL;
  else peakSupra = NULL;
}

void Syllable::removeDuds ()  {
  for (int x = 0; x < onset.size (); x++)
    if (onset[x].isNull ())  {
      onset.removeAt (x);
      onsetSupras.removeAt (x);
      x--;
    }

  for (int x = 0; x < peak.size (); x++)
    if (peak[x].isNull ())  {
      peak.removeAt (x);
      peakSupras.removeAt (x);
      x--;
    }

  for (int x = 0; x < coda.size (); x++)
    if (coda[x].isNull ())  {
      coda.removeAt (x);
      codaSupras.removeAt (x);
      x--;
    }

  if (peakSupra.isNull ())
    peakSupra = NULL;

  if (beforeSupra.isNull ())
    beforeSupra = NULL;

  if (followingSupra.isNull ())
    followingSupra = NULL;

  for (int x = 0; x < onsetSupras.size (); x++)
    for (int y = 0; y < onsetSupras[x].size (); y++)
      if (onsetSupras[x][y].isNull ())  {
        onsetSupras[x].removeAt (y);
        y--;
      }

  for (int x = 0; x < peakSupras.size (); x++)
    for (int y = 0; y < peakSupras[x].size (); y++)
      if (peakSupras[x][y].isNull ())  {
        peakSupras[x].removeAt (y);
        y--;
      }

  for (int x = 0; x < codaSupras.size (); x++)
    for (int y = 0; y < codaSupras[x].size (); y++)
      if (codaSupras[x][y].isNull ())  {
        codaSupras[x].removeAt (y);
        y--;
      }
}

bool Syllable::isEmpty ()  {
  return peak.isEmpty ();
}

QString Syllable::applySupras (QPointer<Phoneme> p,
                               QList< QPointer<Suprasegmental> > s) {
  if (s.isEmpty ()) return p->getName ();

  QString text = "";

  for (int x = 0; x < s.size (); x++)
    if (s[x]->getRepresentationType () == DIACRITIC)
      text = addDiacritic (p->getName (), s[x]->getRepresentationDetail ());

  if (text == "") text += p->getName ();

  for (int x = 0; x < s.size (); x++)
    if (s[x]->getRepresentationType () == DOUBLED)
      text += text;

  for (int x = 0; x < s.size (); x++)  {
    if (s[x]->getRepresentationType () == BEFORE)
      text.prepend (s[x]->getRepresentationDetail ());
    if (s[x]->getRepresentationType () == FOLLOWING)
      text += s[x]->getRepresentationDetail ();
  }

  return text;
}

QString Syllable::addDiacritic (QString str, QString type)  {
  if (type == "Acute Accent")
    for (int x = 0; x < 12; x++)
      str.replace (plainChars[x], diacriticChars[x]);

  if (type == "Grave Accent")
    for (int x = 0; x < 12; x++)
      str.replace (plainChars[x], diacriticChars[x + 12]);

  if (type == "Circumflex")
    for (int x = 0; x < 12; x++)
      str.replace (plainChars[x], diacriticChars[x + 24]);

  if (type == "Diaresis/Umlaut")
    for (int x = 0; x < 12; x++)
      str.replace (plainChars[x], diacriticChars[x + 36]);

  if (type == "Macron")
    for (int x = 0; x < 12; x++)
      str.replace (plainChars[x], diacriticChars[x + 48]);

  return str;
}