#include <QStringListModel>
#include <QLineEdit>
#include <QPushButton>
#include <QListView>
#include <QLabel>
#include <QCheckBox>
#include <QComboBox>
#include <QTextEdit>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QTextStream>

#include "wordpage.h"
#include "word.h"
#include "syllable.h"
#include "editphonologydialog.h"

WordPage::WordPage ()  {
  squareBrackets = false;
  
  wordListModel = new QStringListModel;
  editPhonologyDialog = NULL;

  addWordEdit = new QLineEdit;
  addWordButton = new QPushButton ("Add Word");
  addWordButton->setMaximumSize (addWordButton->sizeHint ());
  deleteButton = new QPushButton ("Delete");
  deleteButton->setMaximumSize (deleteButton->sizeHint ());
  deleteButton->setEnabled (false);
  addWordLayout = new QHBoxLayout;
  addWordLayout->addWidget (addWordButton);
  addWordLayout->addWidget (addWordEdit);
  addWordLayout->addWidget (deleteButton);

  wordListView = new QListView;
  wordListView->setModel (wordListModel);
  wordListView->setSelectionBehavior (QAbstractItemView::SelectRows);
  wordListView->setEditTriggers (QAbstractItemView::NoEditTriggers);
  wordListView->setSelectionMode (QAbstractItemView::ExtendedSelection);

  wordListLayout = new QVBoxLayout;
  wordListLayout->addLayout (addWordLayout);
  wordListLayout->addWidget (wordListView);

  QLabel *wordNameLabel = new QLabel ("Spelling:");
  wordNameEdit = new QLineEdit;
  wordNameLayout = new QHBoxLayout;
  wordNameButton = new QPushButton ("Change");
  wordNameButton->setMaximumSize (wordNameButton->sizeHint ());
  wordNameButton->setEnabled (false);
  wordNameLayout->addWidget (wordNameLabel);
  wordNameLayout->addWidget (wordNameEdit);
  wordNameLayout->addWidget (wordNameButton);

  recalcPhonologyBox =
        new QCheckBox ("Update phonology when spelling is changed");

  QLabel *wordPhonologyLabel = new QLabel ("Phonology:");
  phonologyLabel = new QLabel ("");
  editPhonologyButton = new QPushButton ("Edit");
  editPhonologyButton->setMaximumSize (editPhonologyButton->sizeHint ());
  editPhonologyButton->setEnabled (false);
  phonologyLayout = new QHBoxLayout;
  phonologyLayout->addWidget (wordPhonologyLabel);
  phonologyLayout->addWidget (phonologyLabel);
  phonologyLayout->addWidget (editPhonologyButton);

  QLabel *typeLabel = new QLabel ("Type:");
  typeLabel->setMaximumSize (typeLabel->sizeHint ());
  typeLabel->setMinimumSize (typeLabel->sizeHint ());
  QLabel *subTypeLabel = new QLabel ("Subtype:");
  subTypeLabel->setMaximumSize (subTypeLabel->sizeHint ());
  subTypeLabel->setMinimumSize (subTypeLabel->sizeHint ());
  typeBox = new QComboBox;
  typeBox->setEditable (true);
  typeBox->setEnabled (false);
  typeBox->setSizeAdjustPolicy (QComboBox::AdjustToContents);
  typeBox->setMinimumWidth (100);
  typeLayout = new QHBoxLayout;
  typeLayout->addWidget (typeLabel);
  typeLayout->addWidget (typeBox);
  subTypeBox = new QComboBox;
  subTypeBox->setEditable (true);
  subTypeBox->setEnabled (false);
  subTypeBox->setSizeAdjustPolicy (QComboBox::AdjustToContents);
  subTypeBox->setMinimumWidth (100);
  subTypeLayout = new QHBoxLayout;
  subTypeLayout->addWidget (subTypeLabel);
  subTypeLayout->addWidget (subTypeBox);
  editTypeButton = new QPushButton ("Change Type/Subtype");
  editTypeButton->setMaximumSize (editTypeButton->sizeHint ());
  editTypeButton->setEnabled (false);
  typeBoxesLayout = new QVBoxLayout;
  typeBoxesLayout->addLayout (typeLayout);
  typeBoxesLayout->addLayout (subTypeLayout);
  wordTypeLayout = new QHBoxLayout;
  wordTypeLayout->addLayout (typeBoxesLayout);
  wordTypeLayout->addWidget (editTypeButton);

  connect (typeBox, SIGNAL (currentIndexChanged (const QString)), this,
           SLOT (refreshSubTypeList ()));
  connect (typeBox, SIGNAL (editTextChanged (const QString)), this,
           SLOT (refreshSubTypeList ()));

  QLabel *definitionLabel = new QLabel ("Definition:");
  definitionEdit = new QTextEdit;
  definitionEdit->setEnabled (false);

  wordInfoLayout = new QVBoxLayout;
  wordInfoLayout->addLayout (wordNameLayout);
  wordInfoLayout->addWidget (recalcPhonologyBox);
  wordInfoLayout->addLayout (phonologyLayout);
  wordInfoLayout->addLayout (wordTypeLayout);
  wordInfoLayout->addWidget (definitionLabel);
  wordInfoLayout->addWidget (definitionEdit);

  mainLayout = new QHBoxLayout;
  mainLayout->addLayout (wordListLayout);
  mainLayout->addLayout (wordInfoLayout);

  setLayout (mainLayout);

  connect (addWordButton, SIGNAL (clicked ()), this, SLOT (addWord ()));
  connect (addWordButton, SIGNAL (clicked ()), addWordEdit, SLOT (clear ()));
  connect (deleteButton, SIGNAL (clicked ()), this, SLOT (deleteWord ()));
  connect (wordNameButton, SIGNAL (clicked ()), this, SLOT (editWordName ()));
  connect (editPhonologyButton, SIGNAL (clicked ()), this,
           SLOT (launchEditPhonologyDialog ()));
  connect (editTypeButton, SIGNAL (clicked ()), this, SLOT (editTypes ()));
  connect (definitionEdit, SIGNAL (textChanged ()), this,
           SLOT (editDefinition ()));
}

void WordPage::setModel (QStringListModel *model)  {
  wordListModel = model;
  wordListView->setModel (wordListModel);

  connect (wordListView->selectionModel (),
           SIGNAL (currentChanged (const QModelIndex&, const QModelIndex&)),
           this, SLOT (displayWord ()));
  connect (wordListView->selectionModel (),
           SIGNAL (selectionChanged (const QItemSelection&, const QItemSelection&)),
           this, SLOT (displayWord ()));
}

QList< QPointer<Word> > WordPage::getList ()  {
  return wordList;
}

Phonotactics WordPage::getPhonotactics ()  {
  return phonotactics;
}

void WordPage::setWordList (QList< QPointer<Word> > list)  {
  wordList = list;
  sortWordList ();
  updateModel ();
}

bool WordPage::containsWord (QString word)  {
  for (int x = 0; x < wordList.size (); x++)
    if (wordList[x]->getName () == word)
      return true;

  return false;
}

void WordPage::removeAll (QString word)  {
  for (int x = 0; x < wordList.size (); x++)
    if (wordList[x]->getName () == word)  {
      QPointer<Word> temp = wordList[x];
      wordList.removeAt (x);
      delete temp;
    }
}

void WordPage::setPhonemeList (QList< QPointer<Phoneme> > list)  {
  phonotactics.setPhonemeList (list);

  for (int x = 0; x < wordList.size (); x++)
    wordList[x]->removeDuds ();
}

void WordPage::setAlphabetOrder (QList< QPointer<Phoneme> > list)  {
  alphabetOrder = list;
  sortWordList ();
  updateModel ();
}

void WordPage::setSuprasegmentalList (QList< QPointer<Suprasegmental> > list) {
  phonotactics.setSupraList (list);

  for (int x = 0; x < wordList.size (); x++)
    wordList[x]->removeDuds ();
}

void WordPage::setPhonotactics (Phonotactics phonotax, bool recalc)  {
  phonotax.setSupraList (phonotactics.getSupraList ());
  phonotactics = phonotax;

  if (recalc)  {
    for (int x = 0; x < wordList.size (); x++)
      wordList[x]->setName (wordList[x]->getName (), phonotactics, true);

    sortWordList ();
    updateModel ();
  }
}

void WordPage::setPhonemeFont (QFont font)  {
  phonologyLabel->setFont (font);
}

void WordPage::setOrthographyFont (QFont font)  {
  addWordEdit->setFont (font);
  wordListView->setFont (font);
  wordNameEdit->setFont (font);
}

void WordPage::setNotesFont (QFont font)  {
  typeBox->setFont (font);
  subTypeBox->setFont (font);
  definitionEdit->setFont (font);
}

void WordPage::setSquareBrackets (bool sb)  {
  squareBrackets = sb;

  if (phonologyLabel->text () != "")  {
    QString newText = phonologyLabel->text ();
    newText.chop (1);
    newText.remove (0, 1);

    if (sb) phonologyLabel->setText ("[" + newText + "]");
    else phonologyLabel->setText ("/" + newText + "/");
  }
}

void WordPage::clear ()  {
  phonotactics.clear ();
  wordList.clear ();
  updateModel ();
}

void WordPage::addWords (QList< QPointer<Word> > newWord, bool overwrite)  {
  if (overwrite)  {
    for (int x = 0; x < newWord.size (); x++)
      removeAll (newWord[x]->getName ());
  }

  for (int x = 0; x < newWord.size (); x++)
    wordList.append (newWord[x]);

  sortWordList ();
  updateModel ();
}

void WordPage::addWord ()  {
  QString newWordText = addWordEdit->text ();

  if (newWordText == "") return;

  QPointer<Word> newWord = new Word (newWordText, phonotactics);

  for (int x = 0; x < wordList.size (); x++)
    if (newWord->comesBefore (wordList[x], alphabetOrder))  {
      wordList.insert (x, newWord);
      emit wordListUpdated (wordList);
      updateModel ();
      wordListView->selectionModel ()->setCurrentIndex
                (wordListModel->index (x), QItemSelectionModel::SelectCurrent);
      return;
    }

  int wordIndex = wordList.size ();
  wordList.append (newWord);
  emit wordListUpdated (wordList);
  updateModel ();
  wordListView->selectionModel ()->setCurrentIndex
        (wordListModel->index (wordIndex), QItemSelectionModel::SelectCurrent);
}

void WordPage::deleteWord ()  {
  QModelIndexList indexList = wordListView->selectionModel ()->selectedRows ();

  for (int x = 0; x < indexList.size () - 1; x++)  {
    int largest = x;

    for (int y = x + 1; y < indexList.size (); y++)
      if (indexList[y].row () > indexList[largest].row ())
        largest = y;

    if (largest != x)
      indexList.swap (largest, x);
  }

  for (int x = 0; x < indexList.size (); x++)  {
    QPointer<Word> w = wordList[indexList[x].row ()];
    wordList.removeAt (indexList[x].row ());
    delete w;
  }
  
  emit wordListUpdated (wordList);
  updateModel ();
}

void WordPage::displayWord ()  {
  if (wordListView->selectionModel ()->selectedRows ().size () > 0)
    deleteButton->setEnabled (true);
  else deleteButton->setEnabled (false);

  if (!(wordListView->selectionModel ()->currentIndex ().isValid ()))
    return;
  
  int wordNum = wordListView->selectionModel ()->currentIndex ().row ();
  if (wordList.size () <= wordNum) return;

  wordNameEdit->setEnabled (true);
  wordNameButton->setEnabled (true);
  editPhonologyButton->setEnabled (true);
  typeBox->setEnabled (true);
  subTypeBox->setEnabled (true);
  editTypeButton->setEnabled (true);
  definitionEdit->setEnabled (true);

  wordNameEdit->setText (wordList[wordNum]->getName ());
  if (squareBrackets) phonologyLabel->setText ("[" + wordList[wordNum]->getPhonology () + "]");
  else phonologyLabel->setText ("/" + wordList[wordNum]->getPhonology () + "/");

  refreshTypeList ();
  refreshSubTypeList ();
  definitionEdit->setPlainText (wordList[wordNum]->getDefinition ());
}

void WordPage::editWordName ()  {
  int wordNum = wordListView->selectionModel ()->currentIndex ().row ();
  QPointer<Word> currentWord = wordList[wordNum];

  QString name = wordNameEdit->text ();

  if (name == "") return;

  currentWord->setName(name, phonotactics, recalcPhonologyBox->isChecked ());

  sortWordList ();
  updateModel ();

  QModelIndex index = wordListModel->index (wordList.indexOf (currentWord));
  wordListView->selectionModel ()->setCurrentIndex
                                  (index, QItemSelectionModel::SelectCurrent);
}

void WordPage::launchEditPhonologyDialog ()  {
  int num = wordListView->selectionModel ()->currentIndex ().row ();

  if (editPhonologyDialog) delete editPhonologyDialog;

  editPhonologyDialog = new EditPhonologyDialog
    (wordList[num]->getName (), wordList[num]->getSyllableList (), phonotactics);
  editPhonologyDialog->setPhonemeFont (phonologyLabel->font ());

  connect (editPhonologyDialog, SIGNAL (updatePhonology (QList<Syllable>)), this,
           SLOT (editPhonology (QList<Syllable>)));

  editPhonologyDialog->exec ();
}

void WordPage::editPhonology (QList<Syllable> phonology)  {
  int wordNum = wordListView->selectionModel ()->currentIndex ().row ();
  QPointer<Word> currentWord = wordList[wordNum];
  currentWord->setPhonology (phonology);

  sortWordList ();
  updateModel ();

  QModelIndex index = wordListModel->index (wordList.indexOf (currentWord));
  wordListView->selectionModel ()->setCurrentIndex
                                  (index, QItemSelectionModel::SelectCurrent);
}

void WordPage::editTypes ()  {
  int wordNum = wordListView->selectionModel ()->currentIndex ().row ();
  if (wordNum < 0 || wordNum >= wordList.size ()) return;

  wordList[wordNum]->setType (typeBox->currentText ());
  wordList[wordNum]->setSubType (subTypeBox->currentText ());
}

void WordPage::editDefinition ()  {
  int wordNum = wordListView->selectionModel ()->currentIndex ().row ();
  if (wordNum >= 0 && wordNum < wordList.size ())  {
    wordList[wordNum]->setDefinition (definitionEdit->toPlainText ());
    emit wordListUpdated (wordList);
  }
}

void WordPage::refreshTypeList ()  {
  QStringList types;

  for (int x = 0; x < wordList.size (); x++)
    if (!types.contains (wordList[x]->getType ()))
      types.append (wordList[x]->getType ());

  types.sort ();
  typeBox->clear ();
  typeBox->addItems (types);

  int wordNum = wordListView->selectionModel ()->currentIndex ().row ();
  if (wordNum < 0 || wordNum >= wordList.size ()) return;

  for (int x = 0; x < types.size (); x++)
    if (types[x] == wordList[wordNum]->getType ())
      typeBox->setCurrentIndex (x);
}

void WordPage::refreshSubTypeList ()  {
  QStringList subTypes;

  for (int x = 0; x < wordList.size (); x++)
    if (wordList[x]->getType () == typeBox->currentText () &&
        !subTypes.contains (wordList[x]->getSubType ()))
      subTypes.append (wordList[x]->getSubType ());

  subTypes.sort ();
  subTypeBox->clear ();
  subTypeBox->addItems (subTypes);

  int wordNum = wordListView->selectionModel ()->currentIndex ().row ();
  if (wordNum < 0 || wordNum >= wordList.size ()) return;

  for (int x = 0; x < subTypes.size (); x++)
    if (subTypes[x] == wordList[wordNum]->getSubType ())
      subTypeBox->setCurrentIndex (x);
}

void WordPage::updateModel ()  {
  wordListView->selectionModel ()->reset ();
  wordNameEdit->setText ("");
  wordNameEdit->setEnabled (false);
  wordNameButton->setEnabled (false);
  phonologyLabel->setText ("");
  editPhonologyButton->setEnabled (false);
  typeBox->setEnabled (false);
  typeBox->clear ();
  subTypeBox->setEnabled (false);
  subTypeBox->clear ();
  editTypeButton->setEnabled (false);
  definitionEdit->setEnabled (false);
  definitionEdit->setPlainText ("");

  QStringList nameList;

  for (int x = 0; x < wordList.size (); x++)  {
    if (wordList[x]->getType () != "")
      nameList.append (wordList[x]->getName () + " (" + wordList[x]->getType () + ")");
    else nameList.append (wordList[x]->getName ());
  }

  wordListModel->setStringList (nameList);
}

void WordPage::sortWordList ()  {
  for (int x = 0; x < wordList.size (); x++)  {
    int smallest = x;

    for (int y = x; y < wordList.size (); y++)
      if (wordList[y]->comesBefore (wordList[smallest], alphabetOrder))
        smallest = y;

    if (smallest != x) wordList.swap (x, smallest);
  }
}